
class Residue():
    """
    Residue class is mainly used for plotting a bounding box around the given residue.
    """
    def __init__(self, id=None, chain=0, padding=0):
        """
        Initialize class. Can be empty, but better initialized with residue id and chain id.
        
        :param name: id - Residue id. Count starting at 0. Default: None.
        :param type: int, optional
        :param name: chain - Chain id. Count starting at 0. Specify which chain the residue is on. Default 0 (in case of single chain). Default: 0.
        :param type: int, optional
        :param name: padding - Optional padding to bounding box. Default: 0.
        :param type: int, optional
        """
        if id == None:
            self._id_list = []
            self._chain_list = []
        else:
            self._id_list = [id]
            self._chain_list = [chain]
        self._pad = padding
        
    def add_residue(self, id, chain=0):
        """
        Add a new residue to the internal list of residues.
        
        :param name: id - Residue id. Count starting at 0. Default: None.
        :param type: int, optional
        :param name: chain - Chain id. Count starting at 0. Specify which chain the residue is on. Default 0 (in case of single chain). Default: 0.
        :param type: int, optional
        """
        self._id_list.append(id)
        self._chain_list.append(chain)
    
    def get_res_info(self):
        """
        Returns all internal information of class
        
        :returns: list - list of the current residues
        :returns: list - list of chain ID's corresponding to residues
        :returns: int - padding for bounding box
        """
        return self._id_list, self._chain_list, self._pad
    
    def remove_residue(self, id, chain=0):
        """
        Remove speciefied residue from internal list.

        :param name: id - Residue id. Count starting at 0. Default: None.
        :param type: int, optional
        :param name: chain - Chain id. Count starting at 0. Specify which chain the residue is on. Default 0 (in case of single chain). Default: 0.
        :param type: int, optional
        """
        
        
        curr_idx = self._id_list.index(id)

        len_ = len(self._id_list) 
        while(self._chain_list[curr_idx] != chain and curr_idx < len_):
            # if first instance of residue id does not correspond to the specified chain then search the rest of the list
            curr_idx = curr_idx + 1 + self._id_list[curr_idx + 1:].index(id)
        
        if curr_idx < len_:
            del self._id_list[curr_idx]
            del self._chain_list[curr_idx]
        else: 
            print(f"No such residue found: id {id}, chain: {chain}")
    



"""
https://github.com/bunnech/holoprot/blob/main/holoprot/utils/surface.py is the base for this file. Modifications were made.

Utilities for preparing and computing features on molecular surfaces.
"""
import os
import numpy as np
from numpy.matlib import repmat
from Bio.PDB import PDBParser, Selection
from subprocess import Popen, PIPE
from typing import Tuple, List, Dict
import trimesh 

from provis.utils import RADII, POLAR_HYDROGENS

eps = 1e-6
Surface = Tuple[np.ndarray, np.ndarray, np.ndarray, List[str], Dict[str, str]]


def read_msms(
        file_root: str) -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[str]]:
    """Read surface constituents from output files generated using MSMS.

    :param name: file_root - Root name for loading .face and .vert files (produced by MSMS). Default location is data/tmp/{pdb_id}s
    :param type: str
   
    :returns: numpy.ndarray - vertices
    :returns: numpy.ndarray - faces
    :returns: numpy.ndarray - vertex normals
    :returns: list - List of res_id's from output_pdb_as_xyzrn()
    """
    vertfile = open(file_root + ".vert")
    meshdata = (vertfile.read().rstrip()).split("\n")
    vertfile.close()

    # Read number of vertices.
    count = {}
    header = meshdata[2].split()
    count["vertices"] = int(header[0])
    ## Data Structures
    vertices = np.zeros((count["vertices"], 3))
    normalv = np.zeros((count["vertices"], 3))
    atom_id = [""] * count["vertices"]
    res_id = [""] * count["vertices"]
    for i in range(3, len(meshdata)):
        fields = meshdata[i].split()
        vi = i - 3
        vertices[vi][0] = float(fields[0])
        vertices[vi][1] = float(fields[1])
        vertices[vi][2] = float(fields[2])
        normalv[vi][0] = float(fields[3])
        normalv[vi][1] = float(fields[4])
        normalv[vi][2] = float(fields[5])
        atom_id[vi] = fields[7]
        res_id[vi] = fields[9]
        count["vertices"] -= 1

    # Read faces.
    facefile = open(file_root + ".face")
    meshdata = (facefile.read().rstrip()).split("\n")
    facefile.close()

    # Read number of faces.
    header = meshdata[2].split()
    count["faces"] = int(header[0])
    faces = np.zeros((count["faces"], 3), dtype=int)
    normalf = np.zeros((count["faces"], 3))

    for i in range(3, len(meshdata)):
        fi = i - 3
        fields = meshdata[i].split()
        faces[fi][0] = int(fields[0]) - 1
        faces[fi][1] = int(fields[1]) - 1
        faces[fi][2] = int(fields[2]) - 1
        count["faces"] -= 1

    assert count["vertices"] == 0
    assert count["faces"] == 0

    return vertices, faces, normalv, res_id


def output_pdb_as_xyzrn(pdb_file: str, xyzrn_file: str) -> None:
    """
    Converts a .pdb file to a .xyzrn file.

    :param name: pdb_file - path to PDB File to convert (with extension)
    :param type: str
    :param name: xyzrn_file - path to the xyzrn File (with extension)
    :param type: str
    """
    # get absolute path for xyzrn_file, needed for open
    xyzrn_file = os.path.join(os.getcwd(), xyzrn_file)

    filename = pdb_file.split("/")[-1]
    pdb_id = filename.split(".")[0]

    parser = PDBParser()
    struct = parser.get_structure(id=pdb_id, file=pdb_file)
    outfile = open(xyzrn_file, "w")

    for atom in struct.get_atoms():
        name = atom.get_name()
        residue = atom.get_parent()
        # Ignore hetatms.
        if residue.get_id()[0] != " ":
            continue
        resname = residue.get_resname()
        reskey = residue.get_id()[1]
        chain = residue.get_parent().get_id()
        atomtype = name[0]

        color = "Green"
        coords = None
        if atomtype in RADII and resname in POLAR_HYDROGENS:
            if atomtype == "O":
                color = "Red"
            if atomtype == "N":
                color = "Blue"
            if atomtype == "H":
                if name in POLAR_HYDROGENS[resname]:
                    color = "Blue"  # Polar hydrogens
            coords = "{:.06f} {:.06f} {:.06f}".format(atom.get_coord()[0],
                                                      atom.get_coord()[1],
                                                      atom.get_coord()[2])
            insertion = "x"
            if residue.get_id()[2] != " ":
                insertion = residue.get_id()[2]
            # this is the weird ID; the 9th column of the xyzrn file
            full_id = "{}_{:d}_{}_{}_{}_{}".format(
                chain,
                residue.get_id()[1], insertion, resname, name, color)
        if coords is not None:
            outfile.write(coords + " " + RADII[atomtype] + " 1 " + full_id +
                          "\n")


def get_surface(out_path: str, density: float):
    """
    Wrapper function that reads in the output from the MSMS executable to build the protein surface.

    :param name: out_path - path to output (output path from namechecker) directory. Usually data/tmp
    :param type: str
    :param name: density - Need to pass same density as used by the MSMS binary, as the face and vert files have the density included in their names. The variable is needed for loading these files.
    :param type: bool
        
    :returns: numpy.ndarray - vertices
    :returns: numpy.ndarray - faces
    :returns: numpy.ndarray - vertex normals
    :returns: list - List of res_id's from output_pdb_as_xyzrn()
    :returns: dict - Dictionary: residues as keys, areas as values
    """
    file_base = os.path.abspath(out_path)
    file_base = f"{file_base}_out_{int(density * 10)}"

    # file_base = pdb_path.split(".")[0]
    # file_base = f"{pdb_path}_out_{int(density * 10)}"

    vertices, faces, normals, names = read_msms(file_base)
    areas = {}
    ses_file = open(file_base + ".area")
    next(ses_file)  # ignore header line
    for line in ses_file:
        fields = line.split()
        areas[fields[3]] = fields[1]

    return vertices, faces, normals, names, areas


def compute_normal(vertices: np.ndarray, faces: np.ndarray) -> np.ndarray:
    """
    Compute normals for the vertices and faces

    :param name: vertices - Vertices of the mesh
    :param type: np.ndarray 
    :param name: faces - Faces of the mesh
    :param type: np.ndarray 
    
    :returns: np.ndarray - Normals of the mesh
    """
    vertex = vertices.T
    face = faces.T
    nface = np.size(face, 1)
    nvert = np.size(vertex, 1)
    normal = np.zeros((3, nvert))
    # unit normals to the faces
    normalf = crossp(
        vertex[:, face[1, :]] - vertex[:, face[0, :]],
        vertex[:, face[2, :]] - vertex[:, face[0, :]],
    )
    sum_squares = np.sum(normalf**2, 0)
    d = np.sqrt(sum_squares)
    d[d < eps] = 1
    normalf = normalf / repmat(d, 3, 1)
    # unit normal to the vertex
    normal = np.zeros((3, nvert))
    for i in np.arange(0, nface):
        f = face[:, i]
        for j in np.arange(3):
            normal[:, f[j]] = normal[:, f[j]] + normalf[:, i]

    # normalize
    d = np.sqrt(np.sum(normal**2, 0))
    d[d < eps] = 1
    normal = normal / repmat(d, 3, 1)
    # enforce that the normal are outward
    vertex_means = np.mean(vertex, 0)
    v = vertex - repmat(vertex_means, 3, 1)
    s = np.sum(np.multiply(v, normal), 1)
    if np.sum(s > 0) < np.sum(s < 0):
        # flip
        normal = -normal
        normalf = -normalf
    return normal.T


def crossp(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Creates the cross product of two numpy arrays

    :param name: x - Array 1
    :param type: np.ndarray
    :param name: y - Array 2
    :param type: np.ndarray

    :returns: np.ndarray - (Array 1) x (Array 2)
    """
    z = np.zeros((x.shape))
    z[0, :] = np.multiply(x[1, :], y[2, :]) - np.multiply(x[2, :], y[1, :])
    z[1, :] = np.multiply(x[2, :], y[0, :]) - np.multiply(x[0, :], y[2, :])
    z[2, :] = np.multiply(x[0, :], y[1, :]) - np.multiply(x[1, :], y[0, :])
    return z

def prepare_trimesh(vertices: np.ndarray,
                 faces: np.ndarray,
                 normals: np.ndarray = None,
                 resolution: float = 1.0,
                 apply_fixes: bool = False):
    """
    Prepare the mesh surface given vertices and faces. Optionally, compute
    normals and apply fixes to mesh.

    :param name: vertices - Surface vertices
    :param type: np.ndarray
    :param name: faces - Triangular faces on the mesh
    :param type: np.ndarray
    :param name: normals - Normals for each vertex
    :param type: np.ndarray
    :param name: apply_fixes - Optional application of fixes to mesh. Check fix_mesh for details on fixes. Default: False,
    :param type: bool

    :returns: trimesh.Trimesh - Mesh
    """
    tmp_mesh = trimesh.Trimesh(vertices, faces)
    
    if apply_fixes:
        tmp_mesh.process(validate=1)

    if apply_fixes or normals is None:
        normals = compute_normal(tmp_mesh.vertices, tmp_mesh.faces)

    mesh = trimesh.Trimesh(tmp_mesh.vertices, tmp_mesh.faces, vertex_normals=normals)
    return mesh



def fix_trimesh(mesh, resolution: float = 1.0):
    """
    Applies a predefined set of fixes to the mesh, and converts it to a
    specified resolution. These fixes include removing duplicated vertices wihin
    a certain threshold, removing degenerate triangles, splitting longer edges to
    a given target length, and collapsing shorter edges.

    :param name: mesh - Mesh
    :param type: trimesh.Trimesh
    :param name: resolution - Maximum size of edge in the mesh
    :param tyoe: float

    :returns: trimesh.Trimesh - mesh with all fixes applied
    """
    target_len = resolution
    mesh, _ = trimesh.remove_duplicated_vertices(mesh, 0.001)

    count = 0
    mesh, __ = trimesh.remove_degenerated_triangles(mesh, 100)
    mesh, __ = trimesh.split_long_edges(mesh, target_len)
    num_vertices = mesh.num_vertices
    while True:
        mesh, __ = trimesh.collapse_short_edges(mesh, 1e-6)
        mesh, __ = trimesh.collapse_short_edges(
            mesh, target_len, preserve_feature=True)
        mesh, __ = trimesh.remove_obtuse_triangles(mesh, 150.0, 100)
        if mesh.num_vertices == num_vertices:
            break

        num_vertices = mesh.num_vertices
        #print("#v: {}".format(num_vertices));
        count += 1
        if count > 10: break

    mesh = trimesh.resolve_self_intersection(mesh)
    mesh, __ = trimesh.remove_duplicated_faces(mesh)
    mesh = trimesh.compute_outer_hull(mesh)
    mesh, __ = trimesh.remove_duplicated_faces(mesh)
    mesh, __ = trimesh.remove_obtuse_triangles(mesh, 179.0, 5)
    mesh, __ = trimesh.remove_isolated_vertices(mesh)
    mesh, _ = trimesh.remove_duplicated_vertices(mesh, 0.001)

    return mesh



from genericpath import exists
import os
import provis

class NameChecker:
    """
    The point of this class is to provide uniform naming and path locations to all the other classes of provis.
    NameChecker has class variables and a class method to return these variables. 
    
    Variables: 
    pdb_name - Full path to the pdb file without the .pdb extension. Usually PROVIS_PATH/data/pdb/{pdb_id}.
    out_path - Full path to the temporary files. The names of all temporary files are derived from this variable. Usually PROVIS_PATH/data/tmp/{pdb_id}.
    base_path - Full path of the provis directory or any directory that has the following directory structure within: {path}/data/data, {path}/data/img, {path}/data/tmp, {path}/data/pdb. 
    """
    pdb_name = ""
    out_path = ""
    base_path = ""
    
    def __init__(self, name, base_path: str=None):
        """
        Checks if the provided name corresponds to an existing pdb file.
        Works if "name" points to a pdb file in the working directory or a pdb file in the default pdb file location: provis/data/pdb.
        Always sets "out_path" to default output location: provis/data/tmp.

        :param name: name - Name and path to location of the pdb file. If simply pdb id (file name without .pdb extension) is provided the default directory - provis/data/pdb - will be searched for the file. If no file found the input is returned.
        :param type: str
        :param name: base_path - Easiest if you leave this alone, clone the provis git repo and run everything in the provis root directory. Path to the base directory with the following file structure: {base_path}/data/data, {base_path}/data/img, {base_path}/data/tmp, {base_path}/data/pdb AND {base_path}/binaries that contains the binaries required to run provis. Defaults to the "provis" packages path.
        :param type: str, optional
        """
        if not base_path:
            path = os.path.dirname(provis.__file__)
            # path points to provis/provis, we only want provis/
            base_path = path[0:-6]
        
        name = name.split(".")[0]
        name_pdb = name + ".pdb"

        unkown_name = name.split("/")[-1]
        id = unkown_name.split(".")[0]

        
        if not (exists(name_pdb)):
            # name = name.split(".")[0]
            default = base_path + "data/pdb/" + id
            default_file = default + ".pdb"
            if exists(default_file):
                name = default
            else:
                print("The provided pdb file could not be found. Please make sure you have it downloaded and provided the correct file path.")


        out_path = base_path + "data/tmp/" 
        out_path += id 
        NameChecker.pdb_name = name 
        NameChecker.out_path = out_path
        NameChecker.base_path = base_path
    
    @classmethod
    def return_all(cls):
        """
        Return all class variables.
        
        :return: str - path to the input pdb id (file without the extension)
        :return: str - path to the output pdb id - data/tmp/{pdb_id}
        :return: str - path to the root of the provis directory. Following file structure HAS TO exist within: {path}/data/data, {path}/data/img, {path}/data/tmp, {path}/data/pdb.
        """
        
        return NameChecker.pdb_name, NameChecker.out_path, NameChecker.base_path



from Bio.PDB import *
from Bio import SeqIO
import pyvista as pv
import numpy as np
from os.path import exists
import pandas as pd
from biopandas.mol2 import PandasMol2
import trimesh

from provis.utils.atminfo import import_atm_size_info, import_atm_mass_info
from provis.utils.bond_parser import bond_parser
from provis.utils.name_checker import NameChecker
from provis.src.processing.file_converter import FileConverter


class DataHandler:
    """
    The 'brain' of provis, when it comes to handling atomic positions. 
    
    This class loads information from a variety of files and creates meshes to be plotted. 
    Upper level classes - eg. StickPoint - have their own AtomHandler objects that do all the work.
    """
    def __init__(self):
        """
        Load structure form pdb file (by parsing file) and save Biopython structure representation of the protein. Load all size and color helper dictionaries.
        
        :param name: name - Name of file to be loaded. Passed to name_checker.check_name() to get usable paths.
        :param type: str
        """
        self._path, self._out_path, self._base_path = NameChecker.return_all()
        parser = PDBParser()
        file_name = self._path + ".pdb"
        self._structure = parser.get_structure(self._path, file_name)
                # these two dictionaries have to be manually created
        # could not find volume information for CYS, HIS, LYS, THR, TYR
        # for above mentioned apporximation by "closest" available
        self._res_size_dict = { "ALA": 2.80, "ARG": 3.77, "ASN": 3.18, "ASP": 1.51,
            "CYS": 2.87, "GLN": 3.37, "GLU": 1.68, "GLY": 2.51, "HIS": 3.65,
            "ILE": 3.43, "LEU": 3.42, "LYS": 3.77, "MET": 3.44, "PHE": 3.65,
            "PRO": 3.13, "SER": 2.87, "THR": 2.87, "TRP": 3.54, "TYR": 3.65,
            "VAL": 3.24, "HOH":1.375, "HEM": 33.43 } #HEM vol 156536


        self._res_color_dict = {"ALA": '#13B6E2', "ARG": '#23DEFE', "ASN": '#0EFF57', "ASP": '#822DD2',
            "CYS": '#A22282', "GLN": '#60C7B0', "GLU": '#6913FE', "GLY": '#8E41D0', "HIS": '#7089FC',
            "ILE": '#0CAFF9', "LEU": '#806769', "LYS": '#8B7928', "MET": '#68D139', "PHE": '#8BA695',
            "PRO": '#9FEBA4', "SER": '#BBD7EB', "THR": '#D1A67A', "TRP": '#F93713', "TYR": '#E5613D',
            "VAL": '#128033', "HOH": 'w', "HEM": 'r'}
        self._atoms_size_dict, self._atoms_color_dict, self._vw_dict = import_atm_size_info(1)


    def get_atoms(self, show_solvent=False):
        """
        Creates a dictionary of atomic coordinates from structure
        
        :param name: show_solvent - If True solvent molecules also added to retrun dictionary. Default: False.
        :param type: bool, optional
        
        :return: dict - Dictionary of atomic coordinates by atom type.
        """
        # load file into a python list
        residues = self._structure.get_residues()
        residues_list = list(residues)
        
        # initialize dictionary, then fill up iteratively
        atom_data = dict()
        for residue in residues_list:
            if residue.get_resname() == "HOH" and not show_solvent:
                continue
            for atom in residue:
                type_name = atom.element
                # If atom not in dictionary, add it as key with coords in list
                if type_name not in atom_data:
                    atom_data[type_name] = [atom.get_coord()]
                # If atom already in dictionary, append its coordinates to list
                else:
                    atom_data[type_name].append(atom.get_coord())

        # return the 3D positions of atoms organized by atom type
        return atom_data
        
            
    def get_atoms_IDs(self):
        """
        Get atomic coordinates and residue IDs (format from output_pdb_as_xyzrn()) from the xyzrn file.

        :returns: 
        """
        xyzrnfile = open(self._out_path + ".xyzrn")
        meshdata = (xyzrnfile.read().rstrip()).split("\n")
        xyzrnfile.close()
        
        lenm = len(meshdata)
        res_id = [""] * lenm
        
        atom_data = dict()
        for vi in range(lenm):
            fields = meshdata[vi].split()
            vertices = [None] * 3
            vertices[0] = float(fields[0])
            vertices[1] = float(fields[1])
            vertices[2] = float(fields[2])
            res_id[vi] = fields[5]
            res = res_id[vi].split("_")
            atmtype = res[4][0]
            if atmtype not in atom_data:
                atom_data[atmtype] = [vertices]
            # If atom already in dictionary, append its coordinates to list
            else:
                atom_data[atmtype].append(vertices)


        # return the 3D positions of atoms organized by atom type
        return atom_data, res_id

    def get_residues(self):
        """
        Creates a dictionary of coordinates by residues from structure object
        
        :return: dict - Dictionary of atomic coordinates by residue type.
        """

        # load file into a python list
        residues = self._structure.get_residues()
        res_list = list(residues)

        # initialize dictionary, then fill up iteratively
        res_data = dict()
        for res in res_list:
            type_name = res.get_resname()
            center_coord = [0., 0., 0.]
            c = 0.
            for atom in res:
                center_coord = center_coord + atom.get_coord()
                c += 1.
            center_coord = center_coord / c
            if type_name not in res_data:
                res_data[type_name] = [center_coord]
            else:
                res_data[type_name].append(center_coord)

        # return the 3D positions of atoms organized by atom type
        return res_data

    def get_residue_info(self, res, chain, option):
        """
        Calculates information about specified residue from mol2 file
        
        :param name: res - Residue number of specified residue be looked at.
        :param type: str
        :param name: chain - Chain number of corresponding to residue be looked at.
        :param name: option - choose what property of residue you want. com for Centre Of Mass, ch for charge
        :param type: str - options: com, ch
        
        :return: list - List of COM coords of given (exact) residue.
        """
        
        # load info for given residue
        fname = self._out_path + ".mol2"
        pmol = PandasMol2().read_mol2(fname)
        my = pmol.df[pmol.df['subst_id'] == res]
        
        if len(my) == 0:
            print("ERROR: No such residue: ", res)
            return 1
        
        fid = my.iloc[0]['atom_id'] -1
        curr_chain = 0
        actual = pd.DataFrame(columns = list(my))
        for i, row in my.iterrows():
            if row['atom_id'] != fid + 1:
                curr_chain += 1
            if curr_chain == chain:
                actual = actual.append(row)
            fid = row['atom_id']
                
        # if com option calculate Centre of mass
        if option.upper() == "COM":
            COM_sum = [0.0, 0.0, 0.0]
            mass_dict = import_atm_mass_info()
            a = 0
            for i in range(len(actual)):
                m = mass_dict[actual.iloc[i]['atom_type'][0]]
                COM_sum[0] += actual.iloc[i]['x'] * m
                COM_sum[1] += actual.iloc[i]['y'] * m
                COM_sum[2] += actual.iloc[i]['z'] * m
                a += m
            
            COM = [i/a for i in COM_sum]
            return COM
            
        # if charge option calculate residue charge
        elif option.upper() == "CH" or option.upper() == "CHARGE":
            charge = 0.0
            for i in range(len(my)):
                charge += my.iloc[i]['charge']

            return charge
        
    def get_structure(self):
        """
        Return the loaded structure object
        
        :return: structure
        """
        return self._structure
        
    
    def get_atom_mesh(self, atom_data, vw=0, probe=0, phi_res=10, theta_res = 10):
        """
        Create a list of Shperes and colors representing each atom for plotting. Can later be added to a mesh for plotting.
        
        :param name: atom_data - Dictionary of atoms and their coordinates, by atom type.
        :param type: dict
        :param name: vw, optional - When set to True Van-der-Waals atomic radii used instead of empirical radii. Default: False.
        :param type: bool, optional
        :param name: probe - size of probe (representing the solvent size) needed for surface calculation. Default: 0.
        :param type: int, optional
        :param name: phi_res - pyvista phi_resolution for Sphere objects representing atoms. Default: 10.
        :param type: int, optional
        :param name: theta_res - pyvista theta_resolution for Sphere objects representing atoms. Default: 10.
        :param type: int, optional
        
        :return: list - List of pyvista Shperes representing each atom
        :return: list - List of colors corresponding to each atom
        """
        # these two dictionaries have to be manually created
        # also, if more/other atoms present in protein it will not work

        # create glyphs (spherical) to represent each atom
        atoms_spheres = []
        colors_spheres = []
        rad = probe / 2.0
        for atoms_type in atom_data:

            # create a mesh with each atoms position
            mesh = pv.PolyData(np.array(atom_data[atoms_type]))
            # place a specific sphere at given position
            if not vw:
                sphere = pv.Sphere(radius=self._atoms_size_dict[atoms_type], phi_resolution=phi_res, theta_resolution=theta_res)
            else:
                sphere = pv.Sphere(radius=(self._vw_dict[atoms_type] + rad), phi_resolution=phi_res, theta_resolution=theta_res)
            glyphs = mesh.glyph(geom=sphere)
            atoms_spheres.append(glyphs)

            try:
                # color the sphere according to 'CPK' standard
                colors_spheres.append(self._atoms_color_dict[atoms_type])
            except KeyError:
                # if atom is unrecognized, color it pink
                colors_spheres.append('#DD77FF')

        # return list of spheres and colors representing each atom
        return atoms_spheres, colors_spheres

    def get_atom_trimesh(self, atom_data, vw=False, probe=0):
        """
        Create a list of shperes and colors representing each atom for plotting in a Trimesh format. Used for feature computation in the surface_handler class.
        
        :param name: atom_data - Dictionary of atoms and their coordinates, by atom type.
        :param type: dict
        :param name: vw, optional - When set to True Van-der-Waals atomic radii used instead of empirical radii. Default: False.
        :param type: bool, optional
        :param name: probe - size of probe (representing the solvent size) needed for surface calculation. Default: 0.
        :param type: int, optional
        
        :return: list - List of pyvista Shperes representing each atom
        :return: list - List of colors for each atom
        """
        # these two dictionaries have to be manually created
        # also, if more/other atoms present in protein it will not work

        # create glyphs (spherical) to represent each atom
        atoms_spheres = []
        colors_spheres = []
        rad = probe / 2.0
        

        for atoms_type in atom_data:

            # create a mesh with each atoms position
            # mesh = trimesh.points.PointCloud()#np.array(atom_data[atoms_type]))
            # print(mesh)
            # place a specific sphere at given position

            spheres = []
            for atom in atom_data[atoms_type]:
                if not vw:
                    spheres.append(trimesh.primitives.Sphere(radius=self._atoms_size_dict[atoms_type], center=atom))
                else:
                    spheres.append(trimesh.primitives.Sphere(radius=(self._vw_dict[atoms_type] + rad), center=atom))

            mesh = trimesh.util.concatenate(spheres)
            atoms_spheres.append(mesh)

            try:
                # color the sphere according to 'CPK' standard
                colors_spheres.append(self._atoms_color_dict[atoms_type])
            except KeyError:
                # if atom is unrecognized, color it pink
                colors_spheres.append('#DD77FF')

        # return list of spheres and colors representing each atom
        return atoms_spheres, colors_spheres

    def get_residue_mesh(self, res_data, phi_res = 25, theta_res = 25):
        """
        Create a list of Shperes and colors representing each residue for plotting
        
        :param name: res_data - Dictionary of residues and their coordinates by residue type.
        :param type: dict
        :param name: phi_res - pyvista phi_resolution for Sphere objects representing atoms. Defaul: 25.
        :param type: int, optional
        :param name: theta_res - pyvista theta_resolution for Sphere objects representing atoms. Default:25.
        :param type: int, optional
        
        :return: list - List of pyvista Shperes representing each residue
        :return: list - List of colors for each residue
        """

        # create glyphs (spherical) to represent each res
        res_spheres = []
        colors_spheres = []
        for res_type in res_data:

            sphere = pv.Sphere(radius=0.0, phi_resolution=phi_res, theta_resolution=theta_res)
            try:
                sphere = pv.Sphere(radius=self._res_size_dict[res_type], phi_resolution=phi_res, theta_resolution=theta_res)
                
                #color the sphere according to 'CPK' standard
                colors_spheres.append(self._res_color_dict[res_type])
                
            except KeyError:
                # color unkown to light purple
                colors_spheres.append('#DD77FF')

            # create a mesh with each residues position
            mesh = pv.PolyData(np.array(res_data[res_type]))

            # place a specific sphere at given position
            glyphs = mesh.glyph(geom=sphere)
            res_spheres.append(glyphs)


        # return list of spheres and colors representing each res
        return res_spheres, colors_spheres

    def get_bond_mesh(self):
        """
        Determine bonds from 3D information
        
        :return: list - List of pyvista lines representing each bond.
        """
        fname = self._out_path + ".mol2"

        # Check if mol2 file exists. If not convert it from pdb
        file_exists = exists(fname)
        fc = FileConverter()
        if not file_exists:
            fc.pdb_to_mol2(self._path)
        pmol = PandasMol2().read_mol2(fname)
        bonds_in = bond_parser(fname)

        # Find endpoints of bonds, create Line() and add to list of lines
        bonds = []
        for i in range(len(bonds_in)):
            my = pmol.df.iloc[int(bonds_in.iloc[i]['atom1'])-1][['x', 'y', 'z']]
            c = pmol.df.iloc[int(bonds_in.iloc[i]['atom2'])-1][['x', 'y', 'z']]
            
            my_coord = my.values.tolist()
            c_coord = c.values.tolist()
            # if bonds_in.iloc[i]['bond_type'] == '1': #TODO: differentiate between different bonds.
            line = pv.Line(my_coord, c_coord, resolution=5)
            # elif bonds_in.iloc[i]['bond_type'] != '1':
            #     line = pv.Text3D('||')
            bonds.append(line)
        
        return bonds
        



import pyvista as pv
import numpy as np
import trimesh

from provis.src.processing.surface_handler import SurfaceHandler
from provis.utils.name_checker import NameChecker

class Surface:
    """
    The Surface class is used to visualize the surface information of the given molecule.
    
    All member functions of the class are a "two in one" function:
    - plot the surface created by the MSMS binary. 
    - create a good approximation of the surface natively with o3d and trimesh and plot it. MSMS does not have to be installed for this option. (It is slightly slower and less precise)
    
    Choose between the two by setting the msms Boolean variable (True corresponding to the MSMS binary option is default)
    
    """
    def __init__(self, dens=None):
        """
        Initialize Surface class with given filename. Creates internal data structures; a DataHandler to extract basic surface information and stores it in self._atmsurf (this is a list of Spheres for each atom roughly equating the Van-der-Waals radius).
        
        :param name: dens - sampling density used in msms binary. Also needed to load the face and vert files, as their (file)names include the density
        :param type: float, optional
        """
        self._path, self._out_path, self._base_path = NameChecker.return_all()
        if dens:
            self._density = dens
        self._sh = SurfaceHandler(dens=dens)

    def plot_surface(self, msms=False, outname=None, feature=None, patch=None):
        """
        Plot the surface of protein.
        
        :param name: msms - If True plot msms binary version of surface. If False plot the native (non-binary) surface. Default: False.
        :param type: bool, optional
        :param name: outname - save image of plot to specified filename. Will appear in data/img/ directory. Default: data/img/{self._out_path}_surface.
        :param type: string, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        mesh, cas = self._sh.return_mesh_and_color(msms, feature, patch)
        
        #plot
        pl = pv.Plotter(lighting=None)
        pl.background_color = 'grey'
        pl.enable_3_lights()

        pl.add_mesh(mesh, scalars=cas, cmap='RdBu', smooth_shading=True, show_edges=False)
        # pl.add_mesh(mesh, color="white", smooth_shading=True, style=style, show_edges=False)        
        
        # save a screenshot
        if not outname:
            new_name = self._out_path.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_surface.png'
        pl.show(screenshot=outname, title="Surface")

    def plot_hydrophob(self, msms=False, outname=None):
        """
        Plot the hydrophobic features of a protein.

        :param name: msms - If True plot msms binary version of surface. If False plot the native (non-binary) surface. Default: False.
        :param type: bool, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_hydrophob.png.
        :param type: string, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        mesh, cas = self._sh.return_mesh_and_color(msms, feature="hydrophob")

        # plot surface with feature visualization
        pl = pv.Plotter()
        pl.add_mesh(mesh, scalars=cas, cmap='RdBu', smooth_shading=True, show_edges=False)
        pl.background_color = 'grey'
        pl.camera_position = 'xy'
        
        # save a screenshot
        if not outname:
            new_name = self._out_path.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_hydrophob.png'
        pl.show(screenshot=outname, title="Hydrophob")

    def plot_shape(self, msms=False, outname=None):
        """
        Plot the shape features of a protein.

        :param name: msms - If True plot msms binary version of surface. If False plot the native (non-binary) surface. Default: False.
        :param type: bool, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_shape.png.
        :param type: string, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        
        mesh, cas = self._sh.return_mesh_and_color(msms, feature="shape")

        # plot surface with feature visualization
        pl = pv.Plotter()
        pl.add_mesh(mesh, scalars=cas, cmap='RdBu', smooth_shading=True, show_edges=False)
        pl.background_color = 'grey'
        pl.camera_position = 'xy'
        
        # save a screenshot
        if not outname:
            new_name = self._out_path.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_shape.png'
        pl.show(screenshot=outname, title="Shape")

    def plot_charge(self, msms=False, outname=None):
        """
        Plot the charge features of a protein.

        :param name: msms - If True plot msms binary version of surface. If False plot the native (non-binary) surface. Default: False.
        :param type: bool, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_charge.png.
        :param type: string, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        
        mesh, cas = self._sh.return_mesh_and_color(msms, feature="charge")

        # plot surface with feature visualization
        pl = pv.Plotter()
        pl.add_mesh(mesh, scalars=cas, cmap='RdBu', smooth_shading=True, show_edges=False)
        pl.background_color = 'grey'
        pl.camera_position = 'xy'
        
        # save a screenshot
        if not outname:
            new_name = self._out_path.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_charge.png'
        pl.show(screenshot=outname, title="Charge")


"""
Utility functions for computing hydrogen bonds and electrostatics on protein surface.
"""

import numpy as np
from Bio.PDB import NeighborSearch, calc_angle, calc_dihedral
from Bio.PDB import Vector
from Bio.PDB.Residue import Residue
import numpy as np
from sklearn.neighbors import KDTree

from provis.utils import POLAR_HYDROGENS, ACCEPTOR_ANGLES, HBOND_STD_DEV
from provis.utils import ACCEPTOR_PLANES, RADII, DONOR_ATOMS


def compute_hbond_helper(atom_name: str, res: Residue, v: np.ndarray) -> float:
    """
    Helper function. Computes the charge of the vertex
    
    :param name: atom_name - Name of atom.
    :param type: str
    :param name: res - Residue corresponding to atom.
    :param type: Residue
    :param name: v - Vertices.
    :param type: np.ndarray

    :returns: float - charge of the vertex
    """
    # Check if it is a polar hydrogen.
    if is_polar_hydrogen(atom_name, res.get_resname()):
        donor_atom_name = DONOR_ATOMS[atom_name]
        a = res[donor_atom_name].get_coord()  # N/O
        b = res[atom_name].get_coord()  # H
        # Donor-H is always 180.0 degrees, = pi
        angle_deviation = compute_angle_deviation(a, b, v, np.pi)
        angle_penalty = compute_angle_penalty(angle_deviation)
        return 1.0 * angle_penalty
    # Check if it is an acceptor oxygen or nitrogen
    elif is_acceptor_atom(atom_name, res):
        acceptor_atom = res[atom_name]
        b = acceptor_atom.get_coord()
        a = res[ACCEPTOR_ANGLES[atom_name]].get_coord()
        # 120 degress for acceptor
        angle_deviation = compute_angle_deviation(a, b, v, 2 * np.pi / 3)
        # TODO: This should not be 120 for all atoms, i.e. for HIS it should be
        #       ~125.0
        angle_penalty = compute_angle_penalty(angle_deviation)
        plane_penalty = 1.0
        if atom_name in ACCEPTOR_PLANES:
            try:
                d = res[ACCEPTOR_PLANES[atom_name]].get_coord()
            except:
                return 0.0
            plane_deviation = compute_plane_deviation(d, a, b, v)
            plane_penalty = compute_angle_penalty(plane_deviation)
        return -1.0 * angle_penalty * plane_penalty
        # Compute the
    return 0.0


# Compute the absolute value of the deviation from theta
def compute_angle_deviation(a: np.ndarray, b: np.ndarray, c: np.ndarray,
                            theta: float) -> float:
    """
    Computes the absolute angle deviation from theta. a, b, c form the three
    points that define the angle.

    :param name: a - Coordinate vector of the first point.
    :param type: np.ndarray,
    :param name: b - Coordinate vector of the second point.
    :param type: np.ndarray,
    :param name: c - Coordinate vector of the third point.
    :param type: np.ndarray,
    :param name: theta - Angle to compute deviation with respect to.
    :param type: float
    
    :returns: float - absolute deviation of the angle formed by a, b, c with theta
    """
    return abs(calc_angle(Vector(a), Vector(b), Vector(c)) - theta)


def compute_plane_deviation(a: np.ndarray, b: np.ndarray, c: np.ndarray,
                            d: np.ndarray) -> float:
    """
    Computes the absolute plane deviation from theta. a, b, c form the three
    points that define the angle.

 
    :param name: a - Coordinate vector of the first point.
    :param type: np.ndarray,
    :param name: b - Coordinate vector of the second point.
    :param type: np.ndarray,
    :param name: c - Coordinate vector of the third point.
    :param type: np.ndarray,
    :param name: theta - Angle to compute deviation with respect to.
    :param type: float
    
    :returns: float - absolute deviation of the angle formed by a, b, c with theta
    """
    dih = calc_dihedral(Vector(a), Vector(b), Vector(c), Vector(d))
    dev1 = abs(dih)
    dev2 = np.pi - abs(dih)
    return min(dev1, dev2)


# angle_deviation from ideal value. TODO: do a more data-based solution
def compute_angle_penalty(angle_deviation: float) -> float:
    """
    Compute the angle penalty corresponding to angle of deviation.

    :param name: angle_deviation - Angle of deviation.
    :param type: float

    :returns: float - Angle penalty.
    """
    # Standard deviation: HBOND_STD_DEV
    return max(0.0, 1.0 - (angle_deviation / (HBOND_STD_DEV))**2)


def is_polar_hydrogen(atom_name: str, res_name: str) -> bool:
    """Check if the atom in a given residue has polar hydrogens.

    :param name: atom_name - Name of the atom
    :param type: str
    :param name: res_name - Residue name
    :param type: str
    """
    if atom_name in POLAR_HYDROGENS[res_name]:
        return True
    else:
        return False


def is_acceptor_atom(atom_name: str, res: Residue) -> bool:
    """
    Check if atom is acceptor atom.

    :param name: atom_name - Name of the atom
    :param type: str
    :param name: res - The corresponding residue
    :param type: Residue
    
    :returns: bool - True if conditions met
    """
    if atom_name.startswith("O"):
        return True
    else:
        if res.get_resname() == "HIS":
            if atom_name == "ND1" and "HD1" not in res:
                return True
            if atom_name == "NE2" and "HE2" not in res:
                return True
    return False


def compute_satisfied_CO_HN(atoms):
    """
    Compute the list of backbone C=O:H-N that are satisfied. These will be ignored.
    
    :param name: atoms - list of atoms to be checked.
    :param type: BioPython atoms
    
    :returns: set - set of C=O bonds
    :returns: set - set of H-N bonds
    """
    ns = NeighborSearch(atoms)
    satisfied_CO = set()
    satisfied_HN = set()
    for atom1 in atoms:
        res1 = atom1.get_parent()
        if atom1.get_id() == "O":
            neigh_atoms = ns.search(atom1.get_coord(), 2.5, level="A")
            for atom2 in neigh_atoms:
                if atom2.get_id() == "H":
                    res2 = atom2.get_parent()
                    # Ensure they belong to different residues.
                    if res2.get_id() != res1.get_id():
                        # Compute the angle N-H:O, ideal value is 180 (but in
                        # helices it is typically 160) 180 +-30 = pi
                        angle_N_H_O_dev = compute_angle_deviation(
                            res2["N"].get_coord(),
                            atom2.get_coord(),
                            atom1.get_coord(),
                            np.pi,
                        )
                        # Compute angle H:O=C, ideal value is ~160 +- 20 = 8*pi/9
                        angle_H_O_C_dev = compute_angle_deviation(
                            atom2.get_coord(),
                            atom1.get_coord(),
                            res1["C"].get_coord(),
                            8 * np.pi / 9,
                        )
                        ## Allowed deviations: 30 degrees (pi/6) and 20 degrees
                        #       (pi/9)
                        if (angle_N_H_O_dev - np.pi / 6 < 0 and
                                angle_H_O_C_dev - np.pi / 9 < 0.0):
                            satisfied_CO.add(res1.get_id())
                            satisfied_HN.add(res2.get_id())
    return satisfied_CO, satisfied_HN


def normalize_electrostatics(in_elec: np.ndarray) -> np.ndarray:
    """
    Normalizing charges on the surface, by clipping to upper and lower thresholds
    and converting all values to a -1/1 scale.

    :param name: in_elec - Input charges for all surface vertices
    :param type: np.ndarray
        
    :returns: np.ndarray - Normalized surface vertex charges
    """
    elec = np.copy(in_elec)
    upper_threshold = 3
    lower_threshold = -3
    elec[elec < lower_threshold] = lower_threshold
    elec[elec > upper_threshold] = upper_threshold
    elec = elec - lower_threshold
    elec = elec / (upper_threshold - lower_threshold)
    elec = 2 * elec - 1
    return elec



import pyvista as pv

from provis.src.processing.data_handler import DataHandler
from provis.utils.name_checker import NameChecker

class StickPoint:
    """
    The StickPoint class is used to visualize the structural information of the given molecule. One can easily plot the atoms, residues, bonds or any combination of these.
    """
    def __init__(self):
        """
        Initializes StickPoint class with given filename.
        Creates internal data structures; a DataHandler class instance and loads all the atomic information required for plotting.
        """
        self._id, self._name, self._base_path = NameChecker.return_all()
        # create "brain" of plotting class
        self._dh = DataHandler()
          
        atom_data = self._dh.get_atoms() # second arg: 1 = showsolvent
        ## return list of spheres (meshes) and colors for the spheres
        self._atoms, self._col_a = self._dh.get_atom_mesh(atom_data, vw=0) # second arg: 1 = showvw spheres instead of "normal" radius
        ## return list of lines (meshes)
        self._bonds = self._dh.get_bond_mesh()
        ## return list of spheres (meshes) and colors for the spheres
        res_data = self._dh.get_residues()
        self._residues, self._col_r = self._dh.get_residue_mesh(res_data)
        
        self._atoms_vw, self._col_vw = self._dh.get_atom_mesh(atom_data, vw=1)
        

    def manual_plot(self, atoms=0, col_a=0, box=0, bonds=0, vw=0, residues=0, col_r=0, res=0, outname=0):
        """
        Plot stick and point model
        
        :param name: atoms - List of pyvista Shperes representing each atom, default: 0.
        :param type: list, optional
        :param name: col_a - List of colors for each atom, default: 0.
        :param type: list, optional
        :param name: box - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: bonds - List of pyvista Lines representing each bond, default: 0.
        :param type: list, optional
        :param name: vw - If True styling for Van-der-Waals plotting set. Vw atomic objects still have to be passed under 'atoms' variable.
        :param type: bool, optional
        :param name: res - List of pyvista Shperes representing each residue, default: 0.
        :param type: list, optional
        :param name: col_r - List of colors for each residue, default: 0.
        :param type: list, optional
        :param name: res - Specified residues will be plotted with a bounding box around them.
        :param type: Residue, optional
        :param name: outname - save image of plot to specified filename. Will appear in data/img directory. default: data/img/{self._name}_stick_point.
        :param type: string, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot
        """
        # Use 3 lights because it's a bit brighter
        pl = pv.Plotter(lighting=None)
        pl.background_color = 'grey'
        pl.enable_3_lights()

        # adding the spheres (by atom type) one at a time
        opacity = 1 - vw*0.4
        style = 'surface'
        if atoms != 0:
            if vw:
                style='wireframe'
            for j, mesh in enumerate(atoms):
                pl.add_mesh(mesh, color=col_a[j], opacity=opacity, smooth_shading=True, style=style)

        # adding the bonds one at a time
        if bonds != 0:
            for b in bonds:
                pl.add_mesh(b, color='w', render_lines_as_tubes=True, line_width=5)
        # adding the spheres (by residue) one at a time
        # only executes if residue information provided
        if residues != 0:
            for k, mesh in enumerate(residues):
                pl.add_mesh(mesh, color=col_r[k], opacity=0.2)
        
        # if specified add bounding box
        if box:
            pl.add_bounding_box(color='white', corner_factor=0.5, line_width=1)
        if res:
            res_list, chain_list, pad = res.get_res_info()
            for i, r in enumerate(res_list):
                chain = chain_list[i]
                residues = self._dh.get_structure().get_residues()
                residues_list = list(residues)
                res_name = residues_list[r + 1].get_resname()
                d = (self._dh._res_size_dict[res_name] + pad) * 2
                x, y, z = d,d,d
                pl.add_mesh(pv.Cube(center=self._dh.get_residue_info(r, chain,'com'), x_length=x, y_length=y, z_length=z), style='wireframe', show_edges=1, line_width=5)
        
        # save a screenshot
        if not outname:
            new_name = self._name.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_stick_point.png'
        pl.show(screenshot=outname)


    def plot(self, atoms=0, box=0, bonds=0, vw=0, residues=0, res=None, outname=0, title=None):
        """
        Plot stick and point model
        
        :param name: atoms - Plot atoms, default: 0.
        :param type: bool, optional
        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: bonds, optional - Plot bond, default: 0.
        :param type: bool, optional
        :param name: vw - Plot Wan-der-Waals radii instead of atomic radii.
        :param type: bool, optional
        :param name: residues, optional - Plot residue, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_stick_point.png.
        :param type: string, optional
        :param name: title - Title of the plot window. Defaults to None.
        :param type: str, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot
        """
        # Use 3 lights because it's a bit brighter
        pl = pv.Plotter(lighting=None)
        pl.background_color = 'grey'
        pl.enable_3_lights()

        # adding the spheres (by atom type) one at a time
        opacity = 1 - vw*0.4
        style = 'surface'
        if atoms:
            if vw:
                style='wireframe'
                for j, mesh in enumerate(self._atoms_vw):
                    pl.add_mesh(mesh, color=self._col_vw[j], opacity=opacity, smooth_shading=True, style=style)
            else:
                for j, mesh in enumerate(self._atoms):
                    pl.add_mesh(mesh, color=self._col_a[j], opacity=opacity, smooth_shading=True, style=style)

        # adding the bonds one at a time
        if bonds:
            for b in self._bonds:
                pl.add_mesh(b, color='w', line_width=5, render_lines_as_tubes=True)
                # TODO: figure out why  render_lines_as_tubes=True, crashes
        # adding the spheres (by residue) one at a time
        # only executes if residue information provided
        if residues:
            for k, mesh in enumerate(self._residues):
                pl.add_mesh(mesh, color=self._col_r[k], opacity=0.2)
        
        # if specified add bounding box
        if box:
            pl.add_bounding_box(color='white', corner_factor=0.5, line_width=1)
        
        if res:
            res_list, chain_list, pad = res.get_res_info()
            for i, r in enumerate(res_list):
                chain = chain_list[i]
                residues = self._dh.get_structure().get_residues()
                residues_list = list(residues)
                res_name = residues_list[r + 1].get_resname()
                d = (self._dh._res_size_dict[res_name] + pad) * 2
                x, y, z = d,d,d
                pl.add_mesh(pv.Cube(center=self._dh.get_residue_info(r, chain,'com'), x_length=x, y_length=y, z_length=z), style='wireframe', show_edges=1, line_width=5)

        # save a screenshot
        if not outname:
            new_name = self._name.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_stick_point.png'
        pl.show(screenshot=outname, title=title)


    def plot_stick_point(self, box=0, r=None, outname=0):
        """
        Plot stick and point model
        
        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_stick_point.png.
        :param type: string, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """

        self.plot(atoms=1, box=box, vw=0, bonds=1, residues=0, res=r, outname=outname, title="Stick Point")
        
    def plot_atoms(self, box=0, r=0, outname=0):
        """
        Plot stick and point model
        
        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_atoms.png.
        :param type: string, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        # save a screenshot
        if not outname:
            new_name = self._name.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_atoms.png'
        self.plot(atoms=1, box=box, vw=0, bonds=0, residues=0, res=r, outname=outname, title="Atoms")
        
    def plot_vw(self, box=0, r=0, outname=0):
        """
        Plot Van-der-Waals atoms
        
        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_atoms.png.
        :param type: string, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        # save a screenshot
        if not outname:
            new_name = self._name.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_vw.png'
        self.plot(atoms=1, box=box, vw=1, bonds=0, residues=0, res=r, title="Van-der-Waals")
        
    def plot_bonds(self, box=0, r=0, outname=0):
        """
        Plot bonds only
        
        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_atoms.png.
        :param type: string, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        # save a screenshot
        if not outname:
            new_name = self._name.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_bonds.png'
        self.plot(atoms=0, box=box, vw=0, bonds=1, residues=0, res=r, outname=outname, title="Bonds")




import csv
import os
import provis

def import_atm_size_info(vw=0):
    """
    Funtion to load atomic radii from atmsize.csv, to be used in main
    
    :param name: vw - Option to return vanderwaals radius.
    :param type: bool, optional
    
    :return: dict - dictionary of atomic radius by atom name
    :return: dict - dictionary of color by atom name
    :return: dict - !optional! return dictionary of vw rdius by atom name
    """
    
    path = os.path.dirname(provis.__file__)
    # path points to provis/provis, we only want provis/
    outer_path = path[0:-6]
    
    filename_size = outer_path + "data/data/atmsize.csv"
    filename_vw = outer_path + "data/data/atmvw.csv"
    
    # opening the file using "with" 
    # statement

    firstline = []
    sizedict = {}
    with open(filename_size, newline='') as data:
        firstline = data.readline().rstrip().split(",")
        secondline = data.readline().rstrip().split(",")
        l = len(firstline)
        for i in range(0,l):
            sizedict[firstline[i]] = float(secondline[i])

    color = {  'H': '#FFFFFF',
                        'C': '#333333',
                        'O': '#EA2128',
                        'N': '#2233FF',
                        'S': '#FFDC61',
                        'F': '#1FF01F',
                        'CL': '#1FF01F',
                        'BR': '#992200',
                        'I': '#6600BB',
                        'HE': '#00FFFF',
                        'NE': '#00FFFF',
                        'AR': '#00FFFF',
                        'KR': '#00FFFF',
                        'XE': '#00FFFF',
                        'P': '#FF9900',
                        'B': '#FFAA77',
                        'LI': '#7700FF',
                        'NA': '#7700FF',
                        'K': '#7700FF',
                        'RB': '#7700FF',
                        'CS': '#7700FF',
                        'FR': '#7700FF',
                        'BE': '#007700',
                        'MG': '#007700',
                        'CA': '#007700',
                        'SR': '#007700',
                        'BA': '#007700',
                        'RA': '#007700',
                        'BE': '#007700',
                        'TI': '#999999',
                        'FE': '#DD7700'
                    }
    if vw:
        firstline = []
        vwdict = {}
        with open(filename_vw, newline='') as data:
            firstline = data.readline().rstrip().split(",")
            secondline = data.readline().rstrip().split(",")
            l = len(firstline)
            for i in range(0,l):
                vwdict[firstline[i]] = float(secondline[i])
        return sizedict, color, vwdict
        
    return sizedict, color

# original dictionary used
# atoms_size_dict = {'C': pv.Sphere(radius=0.67, phi_resolution=phi_res, theta_resolution=theta_res),
#                         'O': pv.Sphere(radius=0.48, phi_resolution=phi_res, theta_resolution=theta_res),
#                         'N': pv.Sphere(radius=0.56, phi_resolution=phi_res, theta_resolution=theta_res),
#                         'S': pv.Sphere(radius=0.88, phi_resolution=phi_res, theta_resolution=theta_res),
#                         'CA': pv.Sphere(radius=1.94, phi_resolution=phi_res, theta_resolution=theta_res)
#                         }
# def main():
#     dictt = import_atm_size_info()
#     print(dictt)

# if __name__ == "__main__":
#     main()

  
def import_atm_mass_info():
    """
    Funtion to load atomic mass from atmmass.csv, to be used in main
    
    :return: dict - dictionary of atomic mass by atom name
    """
    path = os.path.dirname(provis.__file__)
    # path points to provis/provis, we only want provis/
    base_path = path[0:-6]
    
    filename_mass = base_path + "data/data/atmmass.csv"
    
    firstline = []
    massdict = {}
    with open(filename_mass, newline='') as data:
        firstline = data.readline().rstrip().split(",")
        secondline = data.readline().rstrip().split(",")
        l = len(firstline)
        for i in range(0,l):
            massdict[firstline[i].upper()] = float(secondline[i])

    return massdict



import numpy as np
import pandas as pd

def bond_parser(filename):
    """
    Calculates bond information from mol2 file.
    
    :param name: filename - Name of file to open
    :param type: str
    
    :returns: pandas.dataFrame - Pandas df with bonds
    """
    f = open(filename,'r')
    f_text = f.read()
    f.close()
    bond_start = f_text.find('@<TRIPOS>BOND')
    bond_end = f_text[bond_start:].replace('@<TRIPOS>BOND','').find('@')
    l = int(len(f_text[bond_start:bond_end].replace('@<TRIPOS>BOND\n','').replace('\n',' ').strip().split())/4)
    df_bonds = pd.DataFrame(np.array(f_text[bond_start:bond_end].replace('@<TRIPOS>BOND\n','').replace('\n',' ').strip().split()).reshape((l,4)), #
            columns=['bond_id', 'atom1', 'atom2', 'bond_type'])
    df_bonds.set_index(['bond_id'], inplace=True)
    return df_bonds





# imports
from genericpath import sameopenfile
import os

import numpy as np
from numpy.lib.twodim_base import tri
import torch
import pyvista as pv
from pyvtk import PolyData, PointData, CellData, Scalars, Vectors, VtkData
import trimesh
from subprocess import PIPE, Popen
import open3d as o3d

from provis.utils.surface_utils import get_surface, compute_normal, prepare_trimesh, fix_trimesh
from provis.utils.name_checker import NameChecker
from provis.src.processing.data_handler import DataHandler
from provis.utils.surface_feat import compute_surface_features

class SurfaceHandler:
    """
    The 'brain' of provis, when it comes to handling surface information. 
    
    This class loads information from a variety of files and creates meshes to be plotted. 
    Upper level classes - eg. StickPoint - have their own AtomHandler objects that do all the work.    
    """
  
    def __init__(self, dens=None):
        """
        Initializes SurfaceHandler class memeber variables: 
        _dh: DataHandler - needed for native mesh creation
        _path: str - path to pdb file (without extension)
        _out_path: str - path to output files (without extension)
        _density: float - density of surface mesh
        _features: tuple(numpy.ndarray) - A collection of arrays representing the feature information for the surface of the protein, saved in a tuple.
        _mesh: Mesh - Surface mesh of protein
        _col: numpy.ndarray - Coloring map of surface. To be passed to pyvista.PolyData.add_mesh(scalars=).

        :param name: dens - Density needed for msms. Defaults to None.
        :param type: float, optional
        """
        self._dh = DataHandler()
        self._path, self._out_path, self._base_path = NameChecker.return_all()
        if dens:
            self._density = dens
        self._features = None
        self._mesh = None
        self._col = None
        
    def get_assignments(self):
        """
        Get assignments (coloring) for the mesh. File has to exist, no way to produce it with provis. 
        Loads provis/data/tmp/{pdb_id}.pth and returns it.

        :returns: PyTourch object - Coloring of surface.
        """
        filename = self._out_path.upper() + '.pth'
        assigment = torch.load(filename)
        
        return assigment

    def get_surface_features(self, mesh, feature):
        """
        Get the coloring corresponding to a specific feature.

        :param name: mesh - The mesh
        :param type: Trimesh
        :param name: feature - Name of feature we are interested in. Options: hydrophob, shape, charge.
        :param type: str
        
        :raises: NotImplementedError - If unkown feature specified error is raised

        returns: numpy.ndarray - Array of coloring corresponding to surface.
        """

        # get surface
        pdb_file = self._path + '.pdb'
        surface = get_surface(self._out_path, density=self._density)

        # compute features of surface
        if not self._features:
            # print(surface)
            # print(mesh)
            self._features = compute_surface_features(surface, pdb_file, self._out_path, mesh, pdb_id = self._path)

        if feature == 'hydrophob':
            return self._features[2]
        elif feature == 'shape':
            return self._features[0]
        elif feature == 'charge':
            return self._features[3]
        else:
            raise NotImplementedError

    def msms_mesh_and_color(self, feature=None, patch=False):
        """
        Return the mesh and coloring ready for plotting.

        :param name: feature - Name of feature, same as in get_surface_features. Options: hydrophob, shape, charge. Defaults to None.
        :param type: str, optional
        :param name: patch - Set coloring of mesh manually, from a file. If set to True get_assignments() will be called. Defaults to False.
        :param type: bool, optional

        :returns: Trimesh - The mesh, always returned
        :returns: numpy.ndarray - Color map corresponding to specified feature. Only returned if a feature is specified as an input. This map can be passed to a pyvista.Plotter.add_mesh() function as the 'scalars' argument to get the encoded coloring.
        """
        
        if not self._mesh:
            surface = get_surface(out_path=self._out_path, density=self._density)
            self._mesh = prepare_trimesh(vertices=surface[0], faces=surface[1], normals=surface[2], 
                            resolution=1.5, apply_fixes=True)
        
        if feature:    
            if patch:
                self._col = self.get_assignments()
            if not patch:
                self._col = self.get_surface_features(self._mesh, feature)

    def native_mesh_and_color(self, feature=None):
        """
        Returns a mesh without the need for the MSMS binary.
        Always returns a pyvista.PolyData mesh of the surface and if a feature is specified it also returns the coloring according to that feature.

        :param name: feature - Name of feature, same as in get_surface_features. Options: hydrophob, shape, charge. Defaults to "".
        :param type: str, optional

        :returns: pyvista.PolyData - The mesh, always returned
        :returns: numpy.ndarray - Color map corresponding to specified feature. Only returned if a feature is specified as an input. This map can be passed to a pyvista.Plotter.add_mesh() function as the 'scalars' argument to get the encoded coloring.
        """
        #old      
        atom_data = self._dh.get_atoms()
        self._atmsurf, col = self._dh.get_atom_mesh(atom_data, vw=1, probe=0.1)
        #old

        # WORKING
        # adding the spheres (by atom type) one at a time
        j = 0
        mesh_ = pv.wrap(self._atmsurf[0])
        for mesh in self._atmsurf[1:]:
            mesh_ = mesh_ + (mesh)

        mesh_ = mesh_.delaunay_3d(alpha=1.5).extract_feature_edges()
        # self._atmsurf, col = self._dh.get_atom_trimesh(atom_data, vw=1, probe=0.1)
        new = trimesh.Trimesh(mesh_.points)#trimesh.util.concatenate(mesh_)
        cloud = o3d.geometry.PointCloud()
        cloud.points = o3d.utility.Vector3dVector(new.vertices)
        cloud.normals = o3d.utility.Vector3dVector(new.vertex_normals)
        radii = [0.1, 0.3, 0.45, 0.6, 0.75, 0.9,1.2, 1.5]
        tri_mesh= o3d.geometry.TriangleMesh.create_from_point_cloud_ball_pivoting(cloud, o3d.utility.DoubleVector(radii))#, depth=depth, width=width, scale=scale, linear_fit=linear_fit)
        v = np.asarray(tri_mesh.vertices)
        f = np.array(tri_mesh.triangles)
        f = np.c_[np.full(len(f), 3), f]
        mesh = pv.PolyData(v, f)
        shell =  mesh.clean().reconstruct_surface()#.clean()
        # WORKING
        shell.compute_normals(inplace=True)
        
        # parse list of PolyData (format) faces and convert them into Trimesh (format) faces 
        len_f = len(shell.faces)
        faces_ = []
        i = 0
        while(i < len_f):
            curr = shell.faces[i]
            temp = [None] * curr
            idxx = 0
            j = i + 1
            while(curr):
                temp[idxx] = shell.faces[j + idxx]
                idxx += 1
                curr -= 1
            faces_.append(temp)
            i += idxx + 1
        
        tri_mesh = trimesh.Trimesh(shell.points, faces=faces_)

        # only needed if feature is specified (-> color map needs to be calculated)
        if feature:
            self._col = self.get_surface_features(tri_mesh, feature)

        self._mesh = tri_mesh
    
    def return_mesh_and_color(self, msms=False, feature=None, patch=False):
        """
        Wrapper function to choose between the msms surface visualization vs the native surface visualization. 
        If you could not download the MSMS binary leave this variable false and you will not run into problems.
        
        :param name: msms - If true, surface generated by msms binary is returned, else the native mesh. Default: False.
        :param type: bool, optional
        :param name: feature - Name of feature, same as in get_surface_features. Options: hydrophob, shape, charge. Defaults to None.
        :param type: str, optional
        :param name: patch - Set coloring of mesh manually. If set to True get_assignments() will be called. Defaults to False.
        :param type: bool, optional
        
        :returns: trimesh.Trimesh - The mesh corresponding to the surface of the protein.
        :returns: numpy.ndarray - Coloring map corresponding to specified feature.
        """
        
        # run appropriate function to compute the mesh and the coloring
        if msms:
            self.msms_mesh_and_color(feature, patch)
        else:
            self.native_mesh_and_color(feature)
          
        return self._mesh, self._col



#!/usr/bin/python3

from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name='provis',
      version='0.0.3',
      description='Protein Visualization Library in Python',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/czirjakkethz/provis',
      author='Kristof Czirjak',
      author_email='czirjakk@student.ethz.ch',
      license='MIT',
      packages=find_packages(),
      install_requires=['biopython', 'trimesh', 'pyvista', 'biopandas', 'torch', 'pyvtk', 'open3d'],
      keywords=['python', 'protein', 'visualization', 'pdb'],
      zip_safe=False,

)



import subprocess
import os.path
from provis.utils.name_checker import NameChecker
from provis.utils.surface_utils import output_pdb_as_xyzrn

class FileConverter():
    """
    Class to create and destroy necessary files required in other parts of code.
    It has a bunch of classmethod functions to call the binaries and scripts to convert the files.
    
    Best practice is to initialize a FileConverter in a given file before calling any function of provis and call the cleanup method after the last plotting function is called. This will keep your directories decluttered. 
    However if you want to plot the same protein many times, then it is benificial to keep the temporary (data/tmp) files as if they exist provis will not recompute them.
    """
    
    _base_path = ""
    def __init__(self, name=None, dens=None, solv=0, bash=0, base_path=None):
        """
        Can be constructed empty. If called with arguments conversions instant. Creates xyzr and mol2 files in every case and face and vert files if msms binary exists.

        :param name: name - Name of file to be loaded. Passed to NameChecker() to get usable paths. Default: None.
        :param type: str, optional
        :param type: float, optional
        :param name: solv - set to True if you want to plot solvent atoms as well. Default: False.
        :param type: bool, optional
        :param name: bash - set to True if you want to run binary version of pdb_to_xyzr (comes with msms). Default: False.
        :param type: bool, optional
        :param name: base_path - Path to "working directory" according to the rules of NameChecker().
        :param name: dens - Density of triangles. Default: None.
        """

        self._solv = solv
        self._bash = bash
        if name:
            NameChecker(name, base_path)
            self._path, self._out_path, FileConverter._base_path = NameChecker.return_all()
            self.pdb_to_mol2(self._path, self._out_path)
            self.pdb_to_pqr(self._path, self._out_path)
            self.pdb_to_xyzrn(self._path, self._out_path)
        if dens:
            self._dens = dens
        if name and dens:
            self.msms(self._out_path, self._dens)
        pass
    
    @classmethod
    def pdb_to_xyzrn(cls, path, output):
        """
        Converts .pdb to .xyzrn file.
        
        :param name: path - Name of input (pdb) file (without extension)
        :param type: str
        :param name: output - Name of output (xyzrn) file (without extension)
        :param type: str
        
        :return: void - xyzrn file
        """

        # create filepath+filename.xyzrn for output file
        xyzrn = output + ".xyzrn"
        name = path + ".pdb"

        if not os.path.exists(xyzrn):
            print("Converting xyzrn file...")
            output_pdb_as_xyzrn(name, xyzrn)
        else:
            print("xyzrn files already exist. No conversion needed")
        
    @classmethod
    def msms(cls, path, dens):
        """
        Run the msms binary for given filename.
        
        It takes the {path}.xyzrn file as input and output is written to {path}_out_{dens}
        Binary path is read in from environment variable: MSMS_BIN. If environment variable does not exist binary will be looked up in provis/binaries/msms.
        
        :param name: path - Path to the/Name of the .xyzrn file to be converted.
        :param type: str
        :param name: dens - Density of triangulation
        :param type: float
        
        :return: void - face and vert files
        """
        # print("./msms.exe -if %4s.xyzr -of %4s_out_%s -density %s" % (name, name, str(int(dens)), str(dens)))

        # Now run MSMS on xyzrn file
        MSMS_BIN = os.environ['MSMS_BIN']
        if not os.path.exists(MSMS_BIN):
            MSMS_BIN = FileConverter._base_path + 'binaries/msms'

        file_base = f"{path}_out_{int(dens * 10)}"

        FNULL = open(os.devnull, 'w')
        args = [MSMS_BIN, "-density", f"{dens}", "-hdensity", "3.0", "-probe",\
                        "1.5", "-if", f"{path}.xyzrn", "-of", file_base, "-af", file_base]
        
        output_exists = os.path.exists(file_base + ".vert") and os.path.exists(file_base + ".face") + os.path.exists(file_base + ".area")
        if os.path.isfile(MSMS_BIN):# and not output_exists:
            if not output_exists:
                print("Converting face and vert files...")
                ac = subprocess.call(args)
            else:
                print("face and vert files already exist. No conversion needed")
        else:
            print("MSMS Binary not found under: ", MSMS_BIN)
            
    @classmethod
    def pdb_to_mol2(cls, path, outpath):
        """
        Run openbabel, to convert pdb to mol2
        
        :param name: path - Name of input (pdb) file (without extension)
        :param type: str        
        :param name: outpath - Name of desired output file (without extension). It will add .mol2 to the given path.
        :param type: str
        """
        
        if not os.path.exists(outpath + ".mol2"):
            print("Converting mol2 file...")
            ac = subprocess.call("obabel %s.pdb -O %s.mol2" % (path,  outpath), shell=True)
        else:
            print("mol2 file already exist. No conversion needed")

    @classmethod
    def pdb_to_pqr(cls, path, outpath, forcefield="swanson"):
        """
        Run pdb2pqr, to convert pdb to pqr
        
        Binary path is read in from environment variable: PDB2PQR_BIN. If environment variable does not exist binary will be looked up in binaries/pdb2pqr/pdb2pqr.
        
        :param name: path - Name of input (pdb) file (without extension)
        :param type: str        
        :param name: outpath - Name of desired output file (without extension). It will add .pqr to the given path.
        :param type: str
        :param name: forcefield - Force field used for charge computation, by binary. Default: swanson. Options: amber, charmm, parse, tyl06, peoepb and swanson
        :param type: str, optional
        """

        PDB2PQR_BIN  = os.environ['PDB2PQR_BIN']
        
        if not os.path.exists(PDB2PQR_BIN):
            PDB2PQR_BIN = FileConverter._base_path + "binaries/pdb2pqr/pdb2pqr"
            
        #possibilities: amber, charmm, parse, tyl06, peoepb and swanson
        args = [PDB2PQR_BIN , f"--ff={forcefield}", f"{path}.pdb", f"{outpath}.pqr"]

        output_exists = os.path.exists(f"{outpath}.pqr")        
        
        if os.path.isfile(PDB2PQR_BIN):# and not output_exists:
            if not output_exists:
                print("Converting pqr file...")
                ac = subprocess.call(args) 
            else:
                print("pqr file already exist. No conversion needed")
        else:
            print("PDB2PQR Binary not found under: ", PDB2PQR_BIN)
       

    @classmethod
    def cleanup(cls, delete_img: bool=False):
        """
        Deletes all files from data/tmp (and data/img) directories.

        :param name: delete_img - If True data/img directory also cleared, else just the data/tmp. Default: False.
        :param type: bool, optional
        """

        import os, shutil
        tmp = FileConverter._base_path + "data/tmp/"
        folders = [tmp]
        if delete_img:
            img = FileConverter._base_path + "data/img/"
            folders.append(img)
        for folder in folders:
            for filename in os.listdir(folder):
                file_path = os.path.join(folder, filename)
                try:
                    if os.path.isfile(file_path) or os.path.islink(file_path):
                        os.unlink(file_path)
                    elif os.path.isdir(file_path):
                        shutil.rmtree(file_path)
                except Exception as e:
                    print('Failed to delete %s. Reason: %s' % (file_path, e))

